/* vi:expandtab:ts=4:sw=4:cindent:tw=80:cinoptions=(0,u0,t0
 *
 * PAF Animation Framework - Clock
 *
 * Copyright © 2008 Fluendo Embedded S.L.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Author(s): Guillaume Emont <guillaume@fluendo.com>
 */

/**
 * SECTION:pafclock
 * @short_description: The object that is responsible for counting time.
 * @see_also: #PafTransformation, #PafTicker.
 *
 * <para>
 * The job of a #PafClock instance is to get ticks (through paf_clock_tick())
 * and call the #PafClock::ticked signal. It provides several options to define
 * how these ticks are transformed into the @fraction parameter of
 * #PafClock::ticked.
 * </para>
 *
 * <refsect2 id="start-pause-stop">
 * <title id="start-pause-stop.title">Start, pause, stop</title>
 * A #PafClock instance can be in three modes:
 * <itemizedlist>
 * <listitem>
 * <emphasis>stopped</emphasis>: it doesn't do anything. This is the mode in
 * which it is after creation.
 * </listitem>
 * <listitem>
 * <emphasis>running</emphasis>: it calls the #PafClock::ticked signal when
 * paf_clock_tick() is called.
 * </listitem>
 * <listitem>
 * <emphasis>paused</emphasis>: it doesn't do anything, but the state in which
 * it was last time it was <emphasis>running</emphasis> is remembered, and will
 * be resumed when the clock is put in <emphasis>running</emphasis> mode again.
 * </listitem>
 * </itemizedlist>
 * </refsect2>
 *
 * <refsect2>
 * <title>Default behaviour</title>
 * If you create a #PafClock instance with paf_clock_new(), and do not change
 * manually any of the properties, the fraction evolves linearly from 0 to 1 in
 * the time of #PafClock:duration.
 * <graphic format="SVG" fileref="../images/clock_default_behaviour.svg"
 * width="50%" />
 * </refsect2>
 *
 * <refsect2 id="repeating">
 * <title id="repeating.title">Repeating</title>
 * <para>
 * A #PafClock can go through #PafClock:duration more (or less) than once if
 * #PafClock:repeat-count is not 1. If it is more than 1, the clock will
 * make other cycles, identical to the first one, with the last one possibly
 * incomplete, if #PafClock::repeat-count is not an integer.
 *
 * Here is what it would look like with the default settings and a
 * #PafClock:repeat-count of 2.5:
 * <graphic format="SVG" fileref="../images/clock_repeat.svg" />
 *
 * If you want your clock to repeat it cycles indefinitely, you should set
 * #PafClock:repeat-count to %PAF_INFINITY.
 *
 * If you set #PafClock:auto-reverse to %TRUE, the clock will do its odd cycles
 * backwards (going from 1.0 to 0.0) as shown in the graphic below.
 * <graphic format="SVG" fileref="../images/clock_auto_reverse.svg" />
 * </para>
 * </refsect2>
 *
 *
 * <refsect2 id="animation-time">
 * <title id="animation-time.title">Animation Time</title>
 * <para>
 * By default (with no options specified), <emphasis>animation time</emphasis>
 * evolves exactly like <emphasis>real time</emphasis>. You can change this by
 * specifying a #PafTransformation that defines the way you want <emphasis>real
 * time</emphasis> to be mapped into <emphasis>animation time</emphasis>.
 * Of course, you can mix these methods together.
 *
 * The <emphasis>animation time</emphasis> is reflected in the @fraction
 * parameter of the #PafClock::ticked signal and is the result of <emphasis>real
 * time</emphasis> on which is applied the transformation set in
 * #PafClock:time-transformation. This allows you to distort time, if you want
 * things to happen faster or slower at some points of the cycle of your clock.
 * For instance, if you want an animation to start accelerating gently, then get
 * to its full speed and eventually to decelerate progressively, you may want to
 * apply a #PafEaseInOutTransformation, which would give you a time curve like
 * this:
 * <graphic format="SVG" fileref="../images/transformation_ease_in_out.svg"
 * width = "50%" />
 * </para>
 * </refsect2>
 */

#include "paf.h"
#include "pafclock.h"

G_DEFINE_TYPE (PafClock, paf_clock, G_TYPE_OBJECT);

#define PAF_CLOCK_GET_PRIVATE(o)\
    (G_TYPE_INSTANCE_GET_PRIVATE ((o), PAF_TYPE_CLOCK, PafClockPrivate))

struct _PafClockPrivate {

    /* Stuff exported as properties */
    guint           duration;
    /* FIXME: find a way to handle repeat_count in a non-float way */
    gfloat          repeat_count;
    gboolean        auto_reverse;
    PafTransformation  *time_transformation;
    PafTicker      *ticker;
    PafState        state;

    /* Über private stuff */
    /* FIXME: add a mutex */
    GTimer         *timer;
    PafModifier    *modifier;
};

/* signals */
enum {
    STATE_CHANGED,
    TICKED,
    LAST_SIGNAL
};

/* property IDs */
enum {
    DURATION=1,
    REPEAT_COUNT,
    AUTO_REVERSE,
    TIME_TRANSFORMATION,
    TICKER,
    STATE,
    LAST_PROPERTY
};

static guint paf_clock_signals[LAST_SIGNAL] = { 0 };

/* Private stuff */

/* Used for the modifier attached to PafClock:time-transformation. We need this
 * since fraction is not a gobject property, but only a parameter to the
 * PafClock::ticked signal. */
static gboolean
paf_clock_fraction_setter (gpointer container, gconstpointer attribute_data,
                           GValue *new_value)
{
    PafClock *clock = PAF_CLOCK (container);
    gfloat fraction = g_value_get_float (new_value);
    gboolean ret = FALSE;

    g_signal_emit (clock, paf_clock_signals[TICKED], 0, fraction, &ret);

    return ret;
}

static void
do_state_update (PafClockPrivate *priv, PafState state)
{
    switch (state)
    {
        case PAF_STATE_STOPPED:
            if (G_LIKELY (priv->timer))
            {
                g_timer_destroy (priv->timer);
                priv->timer = NULL;
            }
            break;
        case PAF_STATE_PAUSED:
            if (G_LIKELY (priv->timer))
            {
                g_timer_stop (priv->timer);
            }
            break;
        case PAF_STATE_RUNNING:
            if (priv->timer)
            {
                g_timer_continue (priv->timer);
            }
            else
            {
                priv->timer = g_timer_new ();
            }
            break;
        default:
            break;
    }
}

/* GObject stuff */

static void
paf_clock_set_property (GObject      *object,
                        guint         property_id,
                        const GValue *value,
                        GParamSpec   *pspec)
{
    PafClock *clock = PAF_CLOCK (object);
    PafClockPrivate *priv = PAF_CLOCK_GET_PRIVATE (clock);

    switch (property_id)
    {
        case DURATION:
            priv->duration = g_value_get_uint (value);
            break;
        case REPEAT_COUNT:
            priv->repeat_count = g_value_get_float (value);
            break;
        case AUTO_REVERSE:
            priv->auto_reverse = g_value_get_boolean (value);
            break;
        case TIME_TRANSFORMATION:
            {
                PafTransformation *transformation = g_value_get_object (value);
                if (priv->time_transformation)
                    g_object_unref (priv->time_transformation);

                if (transformation)
                {
                    g_object_ref (transformation);

                    paf_transformation_add_modifier (transformation,
                                                     priv->modifier);
                }
                priv->time_transformation = transformation;
            }
            break;
        case TICKER:
            {
                PafTicker *ticker = g_value_get_object (value);
                if (priv->ticker)
                {
                    _paf_ticker_remove_clock (priv->ticker, clock);
                    g_object_unref (priv->ticker);
                }

                if (ticker)
                {
                    priv->ticker = g_object_ref (ticker);
                    _paf_ticker_add_clock (priv->ticker, clock);
                }
                else
                    priv->ticker = NULL;
            }
            break;
        case STATE:
            {
                PafState state = g_value_get_enum (value);
                if (state != priv->state)
                {
                    do_state_update (priv, state);
                    priv->state = state;
                    g_signal_emit (clock, paf_clock_signals[STATE_CHANGED],
                                   0, state);
                }
            }
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
    }
}

static void
paf_clock_get_property (GObject      *object,
                        guint         property_id,
                        GValue       *value,
                        GParamSpec   *pspec)
{
    PafClock *clock = PAF_CLOCK (object);
    PafClockPrivate *priv = PAF_CLOCK_GET_PRIVATE (clock);

    switch (property_id)
    {
        case DURATION:
            g_value_set_uint (value, priv->duration);
            break;
        case REPEAT_COUNT:
            g_value_set_float (value, priv->repeat_count);
            break;
        case AUTO_REVERSE:
            g_value_set_boolean (value, priv->auto_reverse);
            break;
        case TIME_TRANSFORMATION:
            g_value_set_object (value, priv->time_transformation);
            break;
        case TICKER:
            g_value_set_object (value, priv->ticker);
            break;
        case STATE:
            g_value_set_enum (value, priv->state);
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
    }
}

static void
paf_clock_dispose (PafClock *self)
{
    PafClockPrivate *priv = PAF_CLOCK_GET_PRIVATE (self);

    priv->duration = 0;
    priv->repeat_count = 0.0f;
    priv->state = PAF_STATE_STOPPED;

    if (priv->time_transformation)
    {
        g_object_unref (priv->time_transformation);
        priv->time_transformation = NULL;
    }

    if (priv->ticker)
    {
        _paf_ticker_remove_clock (priv->ticker, self);
        g_object_unref (priv->ticker);
        priv->ticker = NULL;
    }

    if (priv->timer)
    {
        g_timer_destroy (priv->timer);
        priv->timer = NULL;
    }

    g_object_unref (priv->modifier);

    G_OBJECT_CLASS (paf_clock_parent_class)->dispose ((GObject *) self);
}

static void
paf_clock_init (PafClock *self)
{
    PafClockPrivate *priv;

    priv = PAF_CLOCK_GET_PRIVATE (self);

    priv->time_transformation = NULL;
    priv->ticker = paf_get_default_ticker ();
    priv->state = PAF_STATE_STOPPED;
    priv->timer = NULL;
    priv->modifier = paf_modifier_new_full (paf_clock_fraction_setter,
                                            self, NULL);
}

static void
paf_clock_class_init (PafClockClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);
    GParamSpec *param_spec = NULL;

    g_type_class_add_private (klass, sizeof (PafClockPrivate));

    object_class->set_property = paf_clock_set_property;
    object_class->get_property = paf_clock_get_property;

    /**
     * PafClock:duration:
     *
     * Duration of a cycle of the #PafClock instance in ms.
     *
     * See also <link linkend="repeating" endterm="repeating.title" />.
     */
    param_spec = g_param_spec_uint ("duration", "Duration",
                                    "Duration of a clock cycle in ms",
                                    0, G_MAXUINT, 1000,
                                    G_PARAM_READWRITE | G_PARAM_CONSTRUCT);
    g_object_class_install_property (object_class, DURATION, param_spec);

    /**
     * PafClock:repeat-count:
     *
     * The number of cycles through which the clock must go.
     *
     * See also <link linkend="repeating" endterm="repeating.title" />.
     */
    param_spec = g_param_spec_float ("repeat-count", "Repeat count",
                                    "The number of cycles through which the "
                                    "clock must go",
                                    0.0f, PAF_INFINITY, 1.0f,
                                    G_PARAM_READWRITE | G_PARAM_CONSTRUCT);
    g_object_class_install_property (object_class, REPEAT_COUNT, param_spec);

    /**
     * PafClock:auto-reverse:
     *
     * Whether to reverse time direction when #PafClock:repeat-count is more
     * than 1.
     *
     * See also <link linkend="repeating" endterm="repeating.title" />.
     */
    param_spec = g_param_spec_boolean ("auto-reverse", "Auto reverse",
                                       "Whether to reverse time direction when "
                                       "repeating",
                                       FALSE,
                                       G_PARAM_READWRITE | G_PARAM_CONSTRUCT);
    g_object_class_install_property (object_class, AUTO_REVERSE, param_spec);

    /**
     * PafClock:time-transformation:
     *
     * A #PafTransformation that defines the function that maps time.
     *
     * See also <link linkend="animation-time" endterm="animation-time.title"
     * />.
     */
    param_spec = g_param_spec_object ("time-transformation",
                                      "Time Transformation",
                                      "A PafTransformation that defines the "
                                      "function that maps time",
                                      PAF_TYPE_TRANSFORMATION,
                                      G_PARAM_READWRITE | G_PARAM_CONSTRUCT);
    g_object_class_install_property (object_class,
                                     TIME_TRANSFORMATION, param_spec);

    /**
     * PafClock:ticker:
     *
     * The #PafTicker that updates this clock. If %NULL, the clock will have to
     * be updated by hand with paf_clock_tick().
     *
     * Default value: the default ticker if there is one at the time of creation
     * of the clock (set with paf_set_default_ticker()), %NULL else.
     */
    param_spec = g_param_spec_object ("ticker", "Ticker",
                                      "The PafTicker that updates this clock",
                                      PAF_TYPE_TICKER,
                                      G_PARAM_READWRITE | G_PARAM_CONSTRUCT);
    g_object_class_install_property (object_class, TICKER, param_spec);

    /**
     * PafClock:state:
     *
     * A #PafState saying whether the clock is stopped, running or paused.
     */
    param_spec = g_param_spec_enum ("state", "State",
                                    "The state of the clock",
                                    PAF_TYPE_STATE, PAF_STATE_STOPPED,
                                    G_PARAM_READWRITE);
    g_object_class_install_property (object_class, STATE, param_spec);


    /**
     * PafClock::state-changed:
     * @clock: the #PafClock for which the signal has been emitted.
     * @state: the state (a #PafState) to which the clock has been
     * switched.
     *
     * Will be emitted when the clock's state changes.
     */
    paf_clock_signals[STATE_CHANGED] =
        g_signal_new ("state-changed", G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_CLEANUP, 0, NULL, NULL,
                      paf_marshal_VOID__UINT, G_TYPE_NONE,
                      1, G_TYPE_UINT);

    /**
     * PafClock::ticked:
     * @clock: the #PafClock for which the signal has been emitted.
     * @fraction: the fraction of #PafClock:duration that has elapsed in
     * #PAF_STATE_RUNNING state.
     *
     * Will be emitted at each tick.
     *
     * @returns: whether the signal has had any effect through this callback.
     */
    paf_clock_signals[TICKED] =
        g_signal_new ("ticked", G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_CLEANUP, 0, paf_accumulate_booleans, NULL,
                      paf_marshal_BOOLEAN__FLOAT, G_TYPE_BOOLEAN,
                      1, G_TYPE_FLOAT);

    object_class->dispose = (void (*) (GObject *object)) paf_clock_dispose;

}


/* Public functions */

/**
 * paf_clock_new:
 * @duration: the duration of a cycle of the clock.
 *
 * Creates a new #PafClock instance.
 *
 * Returns: a new #PafClock instance.
 */
PafClock *
paf_clock_new (guint duration)
{
    return g_object_new (PAF_TYPE_CLOCK, "duration", duration, NULL);
}

/**
 * paf_clock_new_full:
 * @duration: the duration of a cycle of the clock, in milliseconds.
 * @repeat_count: the number of cycles the clock will do.
 * @auto_reverse: whether to repeat in auto-reverse mode.
 * @time_transformation is not NULL. Typically one of
 * #PAF_TYPE_EASE_IN_TRANSFORMATION #PAF_TYPE_EASE_OUT_TRANSFORMATION or
 * #PAF_TYPE_EASE_IN_OUT_TRANSFORMATION.
 * @time_transformation: an optional #PafTransformation that specifies a mapping
 * that should be applied on time. You may want to use a
 * #PafEaseInTransformation, a #PafEaseOutTransformation or a
 * #PafEaseInOutTransformation object here.
 * @ticker: a #PafTicker used to synchronize clocks. You probably want the
 * return value of paf_get_default_ticker().
 *
 * Creates a new #PafClock instance.
 *
 * Returns: a new #PafClock instance.
 */
PafClock *
paf_clock_new_full (guint duration,
                    gfloat repeat_count,
                    gboolean auto_reverse,
                    PafTransformation *time_transformation,
                    PafTicker *ticker)
{
    return g_object_new (PAF_TYPE_CLOCK,
                         "duration", duration,
                         "repeat-count", repeat_count,
                         "auto-reverse", auto_reverse,
                         "time-transformation", time_transformation,
                         "ticker", ticker,
                         NULL);
}

/**
 * paf_clock_tick:
 * @clock: a #PafClock instance.
 *
 * This makes @clock emit #PafClock::ticked, with the @fraction parameter
 * calculated according to the properties of @clock. If @clock is not in
 * #PAF_STATE_RUNNING state, this function does nothing.
 *
 * Returns: %TRUE if the tick has triggered an action, %FALSE if it had no
 * effect. In other words, it returns an <emphasis>AND</emphasis> of all the
 * return values of the callback connected to #PafClock::ticked.
 */
/* FIXME: this needs an FPU to be fast */
gboolean
paf_clock_tick (PafClock * clock)
{
    PafClockPrivate *priv;
    gboolean return_value = FALSE;

    g_return_val_if_fail (PAF_IS_CLOCK (clock), FALSE);

    g_object_ref (clock);
    priv = PAF_CLOCK_GET_PRIVATE (clock);

    if (PAF_STATE_RUNNING == priv->state)
    {
        /* FIXME: I'm sure this code could look prettier */
        guint total_elapsed = (1000 * g_timer_elapsed (priv->timer, NULL));
        guint cycle_number = total_elapsed / priv->duration;
        gboolean backwards = priv->auto_reverse ? (cycle_number % 2) : FALSE;
        guint elapsed_in_cycle = total_elapsed % priv->duration;
        gfloat fraction = ((float) elapsed_in_cycle / ((float) priv->duration));

        /* we want to tick for the 1.0f value rather than 0.0f on the next cycle
         * (especially makes sense for repeat_count == 1.0f */
        if (0.0f == fraction && 0 != cycle_number)
        {
            fraction = 1.0f;
            cycle_number--;
        }

        if (cycle_number + fraction > priv->repeat_count)
        {
            /* we are finished here */

            paf_clock_set_state (clock, PAF_STATE_STOPPED);
            goto beach;
        }

        if (backwards)
            fraction = 1.0f - fraction;

        if (priv->time_transformation)
            return_value =
                paf_transformation_update (priv->time_transformation, fraction);
        else
            g_signal_emit (clock, paf_clock_signals[TICKED], 0,
                           fraction, &return_value);
    }

beach:
    g_object_unref (clock);
    return return_value;
}

/**
 * paf_clock_get_duration:
 * @clock: a #PafClock instance
 *
 * Gets the #PafClock:duration property
 */
guint
paf_clock_get_duration (PafClock *clock)
{
    guint duration;

    g_object_get (clock, "duration", &duration, NULL);

    return duration;
}

/**
 * paf_clock_set_duration:
 * @clock: a #PafClock instance
 * @duration: the new duration to set
 *
 * Sets the #PafClock:duration property
 */
void
paf_clock_set_duration (PafClock *clock, guint duration)
{
    g_object_set (clock, "duration", duration, NULL);
}

/**
 * paf_clock_get_repeat_count:
 * @clock: a #PafClock instance
 *
 * Gets the #PafClock:repeat-count property
 */
gfloat
paf_clock_get_repeat_count (PafClock *clock)
{
    gfloat repeat_count;

    g_object_get (clock, "repeat-count", &repeat_count, NULL);

    return repeat_count;
}

/**
 * paf_clock_set_repeat_count:
 * @clock: a #PafClock instance
 * @repeat_count: the new repeat_count to set
 *
 * Sets the #PafClock:repeat-count property
 */
void
paf_clock_set_repeat_count (PafClock *clock, gfloat repeat_count)
{
    g_object_set (clock, "repeat-count", repeat_count, NULL);
}

/**
 * paf_clock_get_auto_reverse:
 * @clock: a #PafClock instance
 *
 * Gets the #PafClock:auto-reverse property
 */
gboolean
paf_clock_get_auto_reverse (PafClock *clock)
{
    gboolean auto_reverse;

    g_object_get (clock, "auto-reverse", &auto_reverse, NULL);

    return auto_reverse;
}

/**
 * paf_clock_set_auto_reverse:
 * @clock: a #PafClock instance
 * @auto_reverse: the new auto_reverse to set
 *
 * Sets the #PafClock:auto-reverse property
 */
void
paf_clock_set_auto_reverse (PafClock *clock,
                            gboolean auto_reverse)
{
    g_object_set (clock, "auto-reverse", auto_reverse, NULL);
}

/**
 * paf_clock_get_time_transformation:
 * @clock: a #PafClock instance
 *
 * Gets the #PafClock:time-transformation property
 */
PafTransformation *
paf_clock_get_time_transformation (PafClock *clock)
{
    PafTransformation *time_transformation;

    g_object_get (clock, "time-transformation", &time_transformation, NULL);

    return time_transformation;
}

/**
 * paf_clock_set_time_transformation:
 * @clock: a #PafClock instance
 * @time_transformation: the new time_transformation to set
 *
 * Sets the #PafClock:time-transformation property
 */
void
paf_clock_set_time_transformation (PafClock *clock,
                                   PafTransformation *time_transformation)
{
    g_object_set (clock, "time-transformation", time_transformation, NULL);
}

/**
 * paf_clock_get_state:
 * @clock: a #PafClock instance
 *
 * Gets the #PafClock:state property
 */
PafState
paf_clock_get_state (PafClock *clock)
{
    PafState state;

    g_object_get (clock, "state", &state, NULL);

    return state;
}

/**
 * paf_clock_set_state:
 * @clock: a #PafClock instance
 * @state: the new state to set
 *
 * Sets the #PafClock:state property
 */
void
paf_clock_set_state (PafClock *clock,
                      PafState state)
{
    g_object_set (clock, "state", state, NULL);
}

/**
 * paf_clock_get_ticker:
 * @clock: a #PafClock instance
 *
 * Gets the #PafClock:ticker property
 */
PafTicker *
paf_clock_get_ticker (PafClock *clock)
{
    PafTicker *ticker;

    g_object_get (clock, "ticker", &ticker, NULL);

    return ticker;
}

/**
 * paf_clock_set_ticker:
 * @clock: a #PafClock instance
 * @ticker: the new ticker to set.
 */
void
paf_clock_set_ticker (PafClock *clock, PafTicker *ticker)
{
    g_object_set (clock, "ticker", ticker, NULL);
}

