/* vi:expandtab:ts=4:sw=4:cindent:tw=80:cinoptions=(0,u0,t0
 *
 * PAF Animation Framework - Ticker
 *
 * Copyright © 2008 Fluendo Embedded S.L.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Author(s): Guillaume Emont <guillaume@fluendo.com>
 */

/**
 * SECTION:pafticker
 * @short_description: Object that tells clocks (#PafClock instances) how to
 * tick.
 * @see_also: #PafClock.
 *
 * <refsect2>
 * A #PafTicker instance maintains a list of clocks (#PafClock instances), and
 * whenever paf_ticker_tick() is called, it calls paf_clock_tick_from_time() for
 * each clock and emits the #PafTicker::ticked signal.
 * </refsect2>
 */

/* FIXME: tests to decide whether to use the PafTicker::ticked signal or to do
 * it by hand.
 */

#include "pafticker.h"

G_DEFINE_TYPE (PafTicker, paf_ticker, G_TYPE_OBJECT);

#define PAF_TICKER_GET_PRIVATE(o)\
    (G_TYPE_INSTANCE_GET_PRIVATE ((o), PAF_TYPE_TICKER, PafTickerPrivate))

struct _PafTickerPrivate {
    GSList *clocks;
    GMutex *clocks_mutex;
};

/* signals */
enum {
    TICKED,
    LAST_SIGNAL
};

static guint paf_ticker_signals[LAST_SIGNAL] = { 0 };

/* private methods */

/* used by paf_ticker_tick() as a parameter of g_slist_foreach() */
static void
do_tick_clock (gpointer _clock, gpointer _return_value)
{
    PafClock *clock = (PafClock *)_clock;
    gboolean *return_value = _return_value, ret=FALSE;

    ret = paf_clock_tick (clock);

    *return_value = *return_value || ret;
}

/* GObject stuff */

static void
paf_ticker_dispose (PafTicker *self)
{
    PafTickerPrivate *priv;

    priv = PAF_TICKER_GET_PRIVATE (self);

    g_mutex_lock (priv->clocks_mutex);
    g_slist_free (priv->clocks);
    priv->clocks = NULL;
    g_mutex_unlock (priv->clocks_mutex);

    G_OBJECT_CLASS (paf_ticker_parent_class)->dispose ((GObject *)self);
}

static void
paf_ticker_finalize (PafTicker *self)
{
    PafTickerPrivate *priv;

    priv = PAF_TICKER_GET_PRIVATE (self);

    g_mutex_free (priv->clocks_mutex);
    priv->clocks_mutex = NULL;
    G_OBJECT_CLASS (paf_ticker_parent_class)->finalize ((GObject *)self);
}

static void
paf_ticker_init (PafTicker *self)
{
    PafTickerPrivate *priv;

    priv = PAF_TICKER_GET_PRIVATE (self);

    priv->clocks = NULL;
    priv->clocks_mutex = g_mutex_new ();
}

static void
paf_ticker_class_init (PafTickerClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    /**
     * PafTicker::ticked:
     * @ticker: the #PafTicker for which the signal has been emitted.
     *
     * Will be emitted at each tick.
     *
     * @returns: whether the signal has had any effect through this callback.
     */
    paf_ticker_signals[TICKED] =
        g_signal_new ("ticked", G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_CLEANUP, 0, paf_accumulate_booleans, NULL,
                      paf_marshal_BOOLEAN__VOID, G_TYPE_BOOLEAN, 0);

    g_type_class_add_private (klass, sizeof (PafTickerPrivate));
    object_class->dispose = (void (*) (GObject *object)) paf_ticker_dispose;
    object_class->finalize = (void (*) (GObject *object)) paf_ticker_finalize;
}

/* Public methods */

/**
 * paf_ticker_new:
 *
 * Creates a new #PafTicker instance.
 *
 * @returns: a new #PafTicker instance.
 */
PafTicker *
paf_ticker_new (void)
{
    return g_object_new (PAF_TYPE_TICKER, NULL);
}

/**
 * paf_ticker_tick:
 * @ticker: a #PafTicker instance.
 *
 * Make @ticker call paf_clock_tick() for each #PafClock instance that it
 * maintains, and emit the #PafTicker::ticked signal.
 *
 * @returns: %TRUE if the tick has triggered an action, %FALSE if it had no effect
 * (in other words, a logical and of the result of paf_clock_tick_from_time()
 * for all the clocks handled by this ticker).
 */
gboolean
paf_ticker_tick (PafTicker *ticker)
{
    gboolean ret = FALSE, signal_return = FALSE;
    PafTickerPrivate *priv;
    GSList *clocks = NULL;

    g_return_val_if_fail (PAF_IS_TICKER (ticker), FALSE);

    g_object_ref (ticker);
    priv = PAF_TICKER_GET_PRIVATE (ticker);

    /* We work on a copy of priv->clocks so that we don't risk interlocking with
     * a function connected to PafClock::ticked. */
    g_mutex_lock (priv->clocks_mutex);
    clocks = g_slist_copy (priv->clocks);
    g_mutex_unlock (priv->clocks_mutex);

    g_slist_foreach (clocks, do_tick_clock, &ret);

    g_signal_emit (ticker, paf_ticker_signals[TICKED], 0, &signal_return);

    g_slist_free (clocks);
    g_object_unref (ticker);
    return ret || signal_return;
}

/**
 * paf_ticker_has_clock:
 * @ticker: a #PafTicker instance.
 * @clock: a #PafClock instance.
 *
 * Check whether @clock is maintained by @ticker.
 *
 * @returns: %TRUE if @clock is maintained by @ticker, %FALSE otherwise.
 */
gboolean
paf_ticker_has_clock   (PafTicker *ticker, const PafClock *clock)
{
    PafTickerPrivate *priv;
    gboolean has_clock = FALSE;

    g_return_val_if_fail (PAF_IS_TICKER (ticker), FALSE);
    if (NULL == clock) return FALSE;
    g_return_val_if_fail (PAF_IS_CLOCK (clock), FALSE);

    g_object_ref (ticker);
    priv = PAF_TICKER_GET_PRIVATE (ticker);

    has_clock = NULL != g_slist_find (priv->clocks, clock);

    g_object_unref (ticker);
    return has_clock;
}

/* public but internal methods */

/**
 * _paf_ticker_add_clock:
 * @ticker: a #PafTicker instance.
 * @clock: a #PafClock instance.
 *
 * Add @clock to the list of #PafClock instances handled by @ticker.
 *
 * This is an internal function and should only be called by the implementation
 * of #PafClock. Use paf_clock_set_ticker ().
 */
void
_paf_ticker_add_clock (PafTicker *ticker, PafClock *clock)
{
    PafTickerPrivate *priv;

    g_return_if_fail (PAF_IS_TICKER (ticker));
    if (NULL == clock) return;
    g_return_if_fail (PAF_IS_CLOCK (clock));

    g_object_ref (ticker);
    priv = PAF_TICKER_GET_PRIVATE (ticker);

    g_mutex_lock (priv->clocks_mutex);
    /* we don't take any reference, we expect the clock to call
     * _paf_ticker_remove_clock() when it is disposed of */
    priv->clocks = g_slist_prepend (priv->clocks, clock);
    g_mutex_unlock (priv->clocks_mutex);

    g_object_unref (ticker);
}

/**
 * _paf_ticker_remove_clock:
 * @ticker: a #PafTicker instance.
 * @clock: a #PafClock instance.
 *
 * Remove @clock from the list of #PafClock instances handled by @ticker.
 *
 * This is an internal function and should only be called by the implementation
 * of #PafClock. Use paf_clock_set_ticker ().
 */
void
_paf_ticker_remove_clock (PafTicker *ticker, PafClock *clock)
{
    PafTickerPrivate *priv;
    GSList *element = NULL;

    g_return_if_fail (PAF_IS_TICKER (ticker));
    if (NULL == clock) return;

    g_object_ref (ticker);
    priv = PAF_TICKER_GET_PRIVATE (ticker);

    g_mutex_lock (priv->clocks_mutex);
    element = g_slist_find (priv->clocks, clock);
    if (element)
    {
        if (NULL != element->data)
        {
            g_object_ref (element->data);
            g_object_unref (element->data);
            element->data = NULL;
        }
        priv->clocks = g_slist_delete_link (priv->clocks, element);
    }
    g_mutex_unlock (priv->clocks_mutex);

    g_object_unref (ticker);
}

