/* vi:expandtab:ts=4:sw=4:cindent:tw=80:cinoptions=(0,u0,t0
 *
 * PAF Animation Framework - Unit tests
 *
 * Copyright © 2008 Fluendo Embedded S.L.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Author(s): Guillaume Emont <guillaume@fluendo.com>
 */

/*
 * FIXME: add stuff to check for g_warning()/g_critical()/g_error()
 */
#include "paftest.h"
#include <string.h>
#include <stdlib.h>
#include <CUnit/Basic.h>
#include <CUnit/Console.h>
#include <CUnit/CUCurses.h>
#include <CUnit/Automated.h>

enum {
    BASIC_INTERFACE=1,
    AUTOMATED_INTERFACE,
    CONSOLE_INTERFACE,
    CURSES_INTERFACE
};

static guint interface_type     = BASIC_INTERFACE;

/* list of GPatternSpecs */
static GSList *valid_test_patterns = NULL;

static gboolean parse_interface (const gchar *option_name,
                                 const gchar *value,
                                 gpointer data,
                                 GError **error);

static GOptionEntry option_entries[] = {
    { "interface", 'i', 0, G_OPTION_ARG_CALLBACK, parse_interface,
        "Use interface_type as interface, one of basic, automated, console, "
        "curses or an unambiguous abreviation of it", "interface_type" },
    { NULL }
};

static gboolean parse_interface (const gchar *option_name,
                                 const gchar *value,
                                 gpointer data,
                                 GError **error)
{
    if (!value)
    {
        g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE,
                     "%s needs an argument", option_name);
    }

    if (value[0] == 'a')
    {
        interface_type = AUTOMATED_INTERFACE;
    }
    else if (value[0] == 'c')
    {
        if (value[1] == 'o')
            interface_type = CONSOLE_INTERFACE;
        else if (value[1] == 'u')
            interface_type = CURSES_INTERFACE;
    }

    /* the default is BASIC_INTERFACE */

    return TRUE;
}

static gint
compare_pattern_string (gconstpointer _pattern, gconstpointer _string)
{
    GPatternSpec *pattern = (GPatternSpec *) _pattern;
    const gchar *string = _string;

    return !g_pattern_match_string (pattern, string);
}

static gboolean
is_test_selected (const char *suite, const char *test)
{
    gchar *full_test_name = g_strjoin (":", suite, test, NULL);
    GSList *found = g_slist_find_custom (valid_test_patterns, full_test_name,
                                         compare_pattern_string);
    g_free (full_test_name);
    return found != NULL;
}

static void
add_test_glob (const char *test_glob)
{
    valid_test_patterns = g_slist_prepend (valid_test_patterns,
                                           g_pattern_spec_new (test_glob));
}

static void
parse_parameter (const char *parameter)
{
    add_test_glob (parameter);
}

static void
parse_parameter_array (int *argc, char ***argv)
{
    int i;

    if (*argc > 1)

    for (i=1; i<*argc; i++)
    {
        parse_parameter ((*argv)[i]);
    }

    *argc=1;
}

static void
parse_options (int *argc, char ***argv)
{
    GError *error = NULL;
    GOptionContext *context =
        g_option_context_new ("[test_suite:[test_name]] ...");

    g_option_context_add_main_entries (context, option_entries, NULL);

    if (!g_option_context_parse (context, argc, argv, &error))
    {
        g_print ("option parsing failed: %s\n", error->message);
        exit (1);
    }

    parse_parameter_array (argc, argv);

    g_option_context_free (context);
}

static void
add_suites (void)
{
    add_pafclock_suite();
    add_pafticker_suite();
}

int
main (int argc, char **argv)
{
    int ret = 0;

    paf_init ();

    parse_options (&argc, &argv);

    CU_initialize_registry ();

    add_suites ();

    switch (interface_type)
    {
        case AUTOMATED_INTERFACE:
            CU_automated_run_tests ();
            break;
        case CONSOLE_INTERFACE:
            CU_console_run_tests ();
            break;
        case CURSES_INTERFACE:
            CU_curses_run_tests ();
            break;
        case BASIC_INTERFACE:
            CU_basic_run_tests ();
            break;
    }

    ret = CU_get_number_of_failures ();

    CU_cleanup_registry ();

    return ret;
}

/* public */
gboolean
paf_test_is_test_selected (const char *suite_name, const char *test_name)
{
    if (!valid_test_patterns) /* no pattern */
        return TRUE;

    return is_test_selected (suite_name, test_name);

}


CU_pSuite
paf_test_add_suite (const char *suite_name,
                    CU_InitializeFunc init,
                    CU_CleanupFunc cleanup)
{
    return CU_add_suite (suite_name, init, cleanup);
}

void
paf_test_add_test (CU_pSuite suite, const char *test_name, CU_TestFunc test)
{
    if (paf_test_is_test_selected (suite->pName, test_name))
        CU_add_test (suite, test_name, test);
}

