/* vi:expandtab:ts=4:sw=4:cindent:tw=80:cinoptions=(0,u0,t0
 *
 * PAF Animation Framework - Unit tests - PafTicker
 *
 * Copyright © 2008 Fluendo Embedded S.L.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Author(s): Guillaume Emont <guillaume@fluendo.com>
 */

/*
 * FIXME:
 *  - test multithreading a bit, that should be done in some stress tests, not
 *  unit tests I guess.
 */

#include "paftest.h"

/* init / cleanup -- not used for construction tests */

static PafTicker *gTicker = NULL;

static int
suite_init (void)
{
    gTicker = paf_ticker_new ();
    if (NULL == gTicker)
        return 1;

    return 0;
}

static int
suite_cleanup (void)
{
    g_object_unref (gTicker);

    return 0;
}

/* construction */

PAF_TEST_CONSTRUCT_DESTRUCT (PafTicker, TICKER, paf_ticker_new ())

/* signals */

static gboolean
ticked_callback (PafTicker *ticker, gboolean *called)
{
    CU_ASSERT_PTR_NOT_NULL_FATAL (called);

    *called = !*called;

    return *called;
}

static void
ticked_signal (void)
{
    gboolean called = FALSE;
    gboolean ret;
    guint handler_id;

    handler_id = g_signal_connect (gTicker, "ticked",
                                   G_CALLBACK (ticked_callback), &called);
    CU_ASSERT_NOT_EQUAL (handler_id, 0);

    CU_ASSERT_FALSE (called);
    ret = paf_ticker_tick (gTicker);
    CU_ASSERT_EQUAL (ret, TRUE);
    CU_ASSERT_EQUAL (called, ret);

    ret = paf_ticker_tick (gTicker);
    CU_ASSERT_EQUAL (ret, FALSE);
    CU_ASSERT_EQUAL (called, ret);

    PAF_TEST_ASSERT_NO_WARNING (g_signal_handler_disconnect (gTicker,
                                                             handler_id));
}

static void
bad_tick (void)
{
    PafClock *clock = NULL;
    PAF_TEST_ASSERT_WARNING (paf_ticker_tick (NULL));

    CU_ASSERT_PTR_NOT_NULL_FATAL (clock = paf_clock_new (1000));
    PAF_TEST_ASSERT_WARNING (paf_ticker_tick ((PafTicker*) clock));

    PAF_TEST_ASSERT_NO_WARNING (g_object_unref (clock));
}

/* general stuff */

static void
bad_add_clock (void)
{
    PafClock *clock;

    PAF_TEST_ASSERT_WARNING (_paf_ticker_add_clock (NULL, NULL));

    PAF_TEST_ASSERT_WARNING (_paf_ticker_add_clock (gTicker, NULL));

    CU_ASSERT_PTR_NOT_NULL_FATAL (clock = paf_clock_new (1000));

    PAF_TEST_ASSERT_WARNING (_paf_ticker_add_clock (NULL, clock));

    PAF_TEST_ASSERT_NO_WARNING (g_object_unref (clock));
}

static void
add_remove_have_clock_parameters (void)
{
    PafClock *clock = NULL;

    CU_ASSERT_PTR_NOT_NULL_FATAL (clock = paf_clock_new (1000));

    PAF_TEST_ASSERT_WARNING (_paf_ticker_add_clock (NULL, NULL));
    PAF_TEST_ASSERT_WARNING (_paf_ticker_add_clock (NULL, clock));

    PAF_TEST_ASSERT_WARNING (paf_ticker_has_clock (NULL, NULL));
    PAF_TEST_ASSERT_WARNING (paf_ticker_has_clock (NULL, clock));

    PAF_TEST_ASSERT_WARNING (_paf_ticker_remove_clock (NULL, NULL));
    PAF_TEST_ASSERT_WARNING (_paf_ticker_remove_clock (NULL, clock));

    PAF_TEST_ASSERT_NO_WARNING (g_object_unref (clock));
}

static void
add_remove_clock (void)
{
    PafClock *clock1 = NULL, *clock2 = NULL;
    gboolean result = FALSE;

    PAF_TEST_ASSERT_NO_WARNING (result = paf_ticker_has_clock (gTicker, NULL));
    CU_ASSERT_FALSE (result);

    CU_ASSERT_PTR_NOT_NULL_FATAL (clock1 = paf_clock_new (1000));
    CU_ASSERT_PTR_NOT_NULL_FATAL (clock2 = paf_clock_new (1000));

    /* no clock */
    PAF_TEST_ASSERT_NO_WARNING (result = paf_ticker_has_clock (gTicker,
                                                                clock1));
    CU_ASSERT_FALSE (result);

    PAF_TEST_ASSERT_NO_WARNING (result = paf_ticker_has_clock (gTicker,
                                                                clock2));
    CU_ASSERT_FALSE (result);

    /* add clock1 */
    PAF_TEST_ASSERT_NO_WARNING (_paf_ticker_add_clock (gTicker, clock1));
    PAF_TEST_ASSERT_NO_WARNING (result = paf_ticker_has_clock (gTicker,
                                                                clock1));
    CU_ASSERT_TRUE (result);
    PAF_TEST_ASSERT_NO_WARNING (result = paf_ticker_has_clock (gTicker,
                                                                clock2));
    CU_ASSERT_FALSE (result);

    /* add clock2 */
    PAF_TEST_ASSERT_NO_WARNING (_paf_ticker_add_clock (gTicker, clock2));
    PAF_TEST_ASSERT_NO_WARNING (result = paf_ticker_has_clock (gTicker,
                                                                clock1));
    CU_ASSERT_TRUE (result);
    PAF_TEST_ASSERT_NO_WARNING (result = paf_ticker_has_clock (gTicker,
                                                                clock2));
    CU_ASSERT_TRUE (result);

    /* remove clock1 */
    PAF_TEST_ASSERT_NO_WARNING (_paf_ticker_remove_clock (gTicker, clock1));
    PAF_TEST_ASSERT_NO_WARNING (result = paf_ticker_has_clock (gTicker,
                                                                clock1));
    CU_ASSERT_FALSE (result);
    PAF_TEST_ASSERT_NO_WARNING (result = paf_ticker_has_clock (gTicker,
                                                                clock2));
    CU_ASSERT_TRUE (result);

    /* remove clock2 */
    PAF_TEST_ASSERT_NO_WARNING (_paf_ticker_remove_clock (gTicker, clock2));
    PAF_TEST_ASSERT_NO_WARNING (result = paf_ticker_has_clock (gTicker,
                                                                clock1));
    CU_ASSERT_FALSE (result);
    PAF_TEST_ASSERT_NO_WARNING (result = paf_ticker_has_clock (gTicker,
                                                                clock2));
    CU_ASSERT_FALSE (result);

    PAF_TEST_ASSERT_NO_WARNING (g_object_unref (clock1));
    PAF_TEST_ASSERT_NO_WARNING (g_object_unref (clock2));
}

/* This checks that tick still works when there's a clock handled. It doesn't
 * check whether the clock's callback is actually called, that should be checked
 * somewhere else, maybe not in unit tests. */
static void
tick_clock (void)
{
    PafClock *clock = NULL;
    gboolean called = FALSE, result=FALSE;
    guint handler_id = 0;

    CU_ASSERT_PTR_NOT_NULL_FATAL (clock = paf_clock_new (1000));

    PAF_TEST_ASSERT_NO_WARNING (_paf_ticker_add_clock (gTicker, clock));
    PAF_TEST_ASSERT_NO_WARNING (result = paf_ticker_has_clock (gTicker, clock));
    CU_ASSERT_TRUE (result);

    PAF_TEST_ASSERT_NO_WARNING (handler_id =
                                g_signal_connect (gTicker, "ticked",
                                                  G_CALLBACK (ticked_callback),
                                                  &called));
    CU_ASSERT_TRUE (0 != handler_id);

    PAF_TEST_ASSERT_NO_WARNING (result = paf_ticker_tick (gTicker));
    CU_ASSERT_TRUE (called);
    CU_ASSERT_TRUE (result);

    PAF_TEST_ASSERT_NO_WARNING (_paf_ticker_remove_clock (gTicker, clock));
    PAF_TEST_ASSERT_NO_WARNING (g_object_unref (clock));
    PAF_TEST_ASSERT_NO_WARNING (g_signal_handler_disconnect (gTicker,
                                                             handler_id));
}

static void
tick_multiple_callbacks (void)
{
    gboolean called = FALSE;
    gboolean ret;
    guint handler_id1, handler_id2;

    handler_id1 = g_signal_connect (gTicker, "ticked",
                                   G_CALLBACK (ticked_callback), &called);
    CU_ASSERT_NOT_EQUAL (handler_id1, 0);

    handler_id2 = g_signal_connect (gTicker, "ticked",
                                   G_CALLBACK (ticked_callback), &called);
    CU_ASSERT_NOT_EQUAL (handler_id2, 0);

    CU_ASSERT_FALSE (called);
    ret = paf_ticker_tick (gTicker);
    CU_ASSERT_EQUAL (ret, TRUE);
    CU_ASSERT_EQUAL (called, FALSE);

    called = TRUE;

    ret = paf_ticker_tick (gTicker);
    CU_ASSERT_EQUAL (ret, TRUE);
    CU_ASSERT_EQUAL (called, TRUE);

    PAF_TEST_ASSERT_NO_WARNING (g_signal_handler_disconnect (gTicker,
                                                             handler_id1));
    PAF_TEST_ASSERT_NO_WARNING (g_signal_handler_disconnect (gTicker,
                                                             handler_id2));
}

void
add_pafticker_suite (void)
{
    CU_pSuite suite = paf_test_add_suite ("PafTicker-noinit", NULL, NULL);

    PAF_TEST_ADD_TEST (suite, construct_destruct);

    suite = paf_test_add_suite ("PafTicker", suite_init, suite_cleanup);
    PAF_TEST_ADD_TEST (suite, ticked_signal);
    PAF_TEST_ADD_TEST (suite, bad_tick);
    PAF_TEST_ADD_TEST (suite, tick_multiple_callbacks);
    PAF_TEST_ADD_TEST (suite, bad_add_clock);
    PAF_TEST_ADD_TEST (suite, add_remove_have_clock_parameters);
    PAF_TEST_ADD_TEST (suite, add_remove_clock);
    PAF_TEST_ADD_TEST (suite, tick_clock);
}

