/* vi:expandtab:ts=4:sw=4:cindent:tw=80:cinoptions=(0,u0,t0
 *
 * This is a list of proof of concept code snipets that often suppose the
 * existence of stuff that doesn't exist. This probably won't compile ever, it
 * is just a tool to test the validity of the API (and not its implementation)
 * as it is developped. Hypothetic stuff have the prefix "magic", possibly with
 * another capitalisation depending on the case.
 */

/**
 * Move @image's x position linearly to @dest. Explicit version.
 */
void
move_image (MagicImage *image, gfloat dest)
{
    PafAnimation *animation;
    PafTransformation *interpolation;

    interpolation = paf_linear_interpolation_new (G_TYPE_FLOAT);
    paf_transformation_add_target (interpolation, image, "x");
    paf_interpolation_add_key_value_from_float (
                                            PAF_INTERPOLATION (interpolation),
                                            0, magic_image_get_x (image), NULL);
    paf_interpolation_add_key_value_from_float (
                                            PAF_INTERPOLATION (interpolation),
                                            1, dest, NULL);

    animation = paf_animation_new (1000.0f, interpolation, NULL);
    paf_animation_start (animation);
}

/**
 * Move @image's x position linearly to @dest. Implicit version.
 */
void
implicitly_move_image (MagicImage *image, gfloat dest)
{
    PafImplicitAnimator *animator = paf_implicit_animator_new (image);

    g_object_set (animator, "x", dest, NULL);
}

/**
 * Move @image's x position smoothly to @dest. Explicit version
 */
void
move_image_smooth (MagicImage *image, gfloat dest)
{
    PafAnimation *animation;
    PafTransformation *interpolation, *easeinout;
    PafClock *clock;

    interpolation = paf_linear_interpolation_new (G_TYPE_FLOAT);
    paf_transformation_add_target (interpolation, image, "x");
    paf_interpolation_add_key_value_from_float (
                                            PAF_INTERPOLATION (interpolation),
                                            0, magic_image_get_x (image), NULL);
    paf_interpolation_add_key_value_from_float (
                                            PAF_INTERPOLATION (interpolation),
                                            1, dest, NULL);

#if method1
    easeinout = paf_ease_in_out_transformation_new ();
    clock = paf_clock_new (2000);
    paf_clock_set_time_transformation (clock, easeinout);

    animation = paf_animation_new_full (clock, interpolation, NULL);
#elif method2
    /* In python, that would give:
     * animation = paf.Animation(2000, interpolation)
     * animation.clock.time_transformation = paf.EaseInOutTransformation()
     */
    animation = paf_animation_new (2000, interpolation, NULL);

    clock = paf_animation_get_clock (animation);
    easeinout = paf_ease_in_out_transformation_new ();
    paf_clock_set_time_transformation (clock, easeinout);
#elif method3
    animation = paf_animation_new (2000, interpolation, NULL);

    clock = paf_animation_get_clock (animation);
    paf_clock_set_time_transformation_type (clock,
                                        PAF_TYPE_EASE_IN_OUT_TRANSFORMATION);
#elif method4
    clock = paf_clock_new_full (2000,
                                1.0f, /* repeat count */
                                FALSE, /* auto_reverse */
                                PAF_TYPE_EASE_IN_OUT_TRANSFORMATION,
                                /*time transformation type*/
                                NULL /* time transformation if G_TYPE_NONE */
                               );

    animation = paf_animation_new_full (clock, interpolation, NULL);
#endif

    paf_animation_start (animation);
}



/**
 * Move @image's x position smoothly to @dest. Implicit version.
 */
void
implicitly_move_image_smooth (MagicImage *image, gfloat dest)
{
    PafImplicitAnimator *animator = paf_implicit_animator_new (image);

    paf_implicit_animator_set_time_transformation_type (PAF_TYPE_EASE_IN_OUT_TRANSFORMATION);

    g_object_set (animator, "x", dest, NULL);
}

/**
 * Make @image move smoothly to @dest, passing through point1, point2 and
 * point3. Implicit version.
 * FIXME: what about expressing time transformations for each step? How could we
 * do that? Should we rely only on the interpolation options to solve the
 * problems that would be solved by time transformations for each step?
 */
void
implicitly_move_through_points (MagicImage *image, MagicPosition *point1,
                                MagicPosition *point2, MagicPosition *point3,
                                MagicPosition *dest)
{
    GSList * positions = NULL;
    PafImplicitAnimator *animator = paf_implicit_animator_new (image);

    paf_implicit_animator_set_time_transformation_type (animator,
                                        PAF_TYPE_EASE_IN_OUT_TRANSFORMATION);

    paf_implicit_animator_set_property_values (animator, "position",
                                               point1, point2, point3, dest,
                                               NULL);
}

/**
 * Make @image move smoothly to @dest, passing through point1, point2 and
 * point3. Explicit version
 */
void
move_through_points (MagicImage *image, MagicPosition *point1,
                     MagicPosition *point2, MagicPosition *point3,
                     MagicPosition *dest)
{
    PafAnimation *animation;
    PafClock *clock;
    PafTransformation *interpolation;
    GValue val1 = {0}, val2 = {0}, val3 = {0}, val_dest = {0};
    PafKeyValue *kval1=NULL, *kval2=NULL, *kval3=NULL, *kval_dest=NULL;

    interpolation = paf_hermite_interpolation_new (MAGIC_TYPE_POSITION);
    paf_transformation_add_target (interpolation, image, "position");

    animation = paf_animation_new (2000, interpolation, NULL);

    /* this would look much better if the values where passed as a list through
     * which we could iterate */
    g_value_init (&val1, MAGIC_TYPE_POSITION);
    g_value_set_instance (&val1, point1);
    kval1 = paf_key_value_new (0.0f, &val1, NULL);

    g_value_init (&val2, MAGIC_TYPE_POSITION);
    g_value_set_instance (&val2, point2);
    kval2 = paf_key_value_new (0.33f, &val2, NULL);

    g_value_init (&val3, MAGIC_TYPE_POSITION);
    g_value_set_instance (&val3, point3);
    kval3 = paf_key_value_new (0.66f, &val3, NULL);

    g_value_init (&val_dest, MAGIC_TYPE_POSITION);
    g_value_set_instance (&val_dest, dest);
    kval_dest = paf_key_value_new (1.0f, &val_dest, NULL);

    paf_interpolation_add_key_values (PAF_INTERPOLATION (interpolation),
                                      kval1, kval2, kval3, kval_dest, NULL);
}

/**
 * See https://code.fluendo.com/pigment/trac/wiki/ExistingTimingFrameworks#Rollercoaster.
 * @positions is a list of MagicPosition instances.
 */
void
roller_coaster_simple (MagicImage *image, GSList *positions)
{
    GSList walk = positions, values = NULL;
    PafImplicitAnimator animator = paf_implicit_animator_new (image);

    while (walk)
    {
        GValue *value = g_malloc0 (sizeof(GValue));

        g_value_init (value, MAGIC_TYPE_POSITION);
        g_value_set_instance (value, walk->data);
        values = g_slist_prepend (values, value);

        walk = walk->next;
    }

    values = g_slist_reverse (values);

    paf_implicit_animator_set_interpolation_type (animator,
                                              PAF_TYPE_CARDINAL_INTERPOLATION);
    paf_implicit_animator_set_property_value_list (animator,
                                                   "position", values);
}

/**
 * See https://code.fluendo.com/pigment/trac/wiki/ExistingTimingFrameworks#Rollercoaster.
 * @positions is a list of MagicPosition, @speeds is a list of gfloats, 0 means
 * the object stops at that point, 1 means "normal" speed at that point, 2 twice
 * as fast, etc... The two lists (@positions and @speeds) must have the same
 * length.
 */
void
roller_coaster (MagicImage *image, GSList *positions, GSList *speeds)
{
    PafAnimation *animation = NULL;
    GSList position = positions, speed = speeds, key_vals = NULL;
    PafTransformation *interpolation =
        paf_cardinal_interpolation_new (MAGIC_TYPE_POSITION);
    paf_transformation_add_target (interpolation, image, "position");

    while (position && speed)
    {
        PafKeyValue *key_val = paf_key_value_new ();
        PafCardinalInterpolationOptions *options =
            paf_cardinal_interpolation_options_new ();

        g_value_init (&key_val->value, MAGIC_TYPE_POSITION);
        g_value_set_instance (&key_val->value, position->data);

        options->speed = speed->data;
        options->ignore_key_time = TRUE;

        key_val->options = options;

        key_vals = g_slist_prepend (key_vals, key_val);

        position = position->next;
        speed = speed->next;
    }
    if (speed || position)
    {
        g_print ("speeds and positions are not lists of the same size\n");
        return;
    }

    key_vals = g_slist_reverse (key_vals);

    paf_interpolation_add_key_value_list (PAF_INTERPOLATION (interpolation),
                                          key_vals);

    animation = paf_animation_new (10000, interpolation, NULL);
    paf_animation_start (animation);
}


/**
 * Does the same as roller_coaster(), using nice wrappers provided by the
 * #PafCardinalInterpolation object.
 */
void
roller_coaster_simpler (MagicImage *image, GSList *positions, GSList *speeds)
{
    PafAnimation *animation = NULL;
    GSList position = positions, speed = speeds;
    PafTransformation interpolation =
        paf_cardinal_interpolation_new (MAGIC_TYPE_POSITION);
    paf_transformation_add_target (interpolation, image, "position");

    while (position && speed)
    {
        GValue value = {0,};

        g_value_init (&value, MAGIC_TYPE_POSITION);
        g_value_set_instance (&value, position->data);

        paf_cardinal_interpolation_add_key_value_full (
                                   PAF_CARDINAL_INTERPOLATION (interpolation),
                                   0.0f, &value, *(float *)speed->data, TRUE);

        position = position->next;
        speed = speed->next;
    }
    if (speed || position)
    {
        g_warning ("speeds and positions are not lists of the same size\n");
        return;
    }

    animation = paf_animation_new (10000, interpolation, NULL);
    paf_animation_start (animation);
}

/**
 * Does exactly what's described at
 * https://code.fluendo.com/pigment/trac/wiki/ExistingTimingFrameworks#Rollercoaster.
 */
void
roller_coaster_predef (MagicImage *image)
{
    gfloat position_x[] = [1.0f, 2.0f, 3.0f, 3.5f, 0.25f, 1.0f];
    gfloat position_y[] = [0.5f, 2.0f, 1.5f, 2.5f, 1.5f,  0.5f];
    gfloat speed[]      = [0.0f, 2.0f, 1.0f, 2.0f, 0.9f,  0.0f];
    gsize table_length  = sizeof(position_x) / sizeof(position_x[0]);
    PafTransformation *interpolation =
        paf_cardinal_interpolation_new (MAGIC_TYPE_POSITION);
    paf_transformation_add_target (interpolation, image, "position");
    gsize i;

    for (i=0; i<table_length; i++)
    {
        MagicPosition *position = magic_position_new (position_x[i],
                                                      position_y[i], 0.0f);
        GValue value = {0};

        g_value_init (value, MAGIC_TYPE_POSITION);
        g_value_set_instance (value, position);

        paf_cardinal_interpolation_add_key_value_full (
                                   PAF_CARDINAL_INTERPOLATION (interpolation),
                                   0.0f, &value, speed[i], TRUE);
    }

    animation = paf_animation_new (10000, interpolation, NULL);
    paf_animation_start (animation);
}

/*
 * Below: code for
 * https://code.fluendo.com/pigment/trac/wiki/ExistingTimingFrameworks#Elisasselectorinteractive
 * An image moves smoothly (ease-in/out) between discreet y positions (by
 * increment of 0.1). The moves are triggered by key presses (up or down), a key
 * press interrupts or changes the current animation, multiple consecutive key
 * presses change the destination and increases the time to reach there.
 */
void
on_key_up (PafImplicitAnimator *animator, gpointer *data)
{
    animator.y -= 0.1f;
}

void
on_key_down (PafImplicitAnimator *animator, gpointer *data)
{
    animator.y += 0.1f;
}

void
setup (MagicImage *image, MagicEventHandler *event_handler)
{
    PafImplicitAnimator *animator = paf_implicit_animator_new (image);

    paf_implicit_animator_set_time_transformation_type (animator,
                                        PAF_TYPE_EASE_IN_OUT_TRANSFORMATION);
    g_signal_connect (event_handler, "key-up",
                      G_CALLBACK (on_key_up), NULL);
    g_signal_connect (event_handler, "key-down",
                      G_CALLBACK (on_key_down), NULL);
}

/*
 * Example of how to integrate PAF in an application using a GMainLoop.
 */
gboolean tick_always_true (gpointer data)
{
    PafTicker *ticker = PAF_TICKER (data);

    paf_ticker_tick (ticker);

    return TRUE;
}

void
my_application_init (guint *source_id)
{
    PafTicker *ticker = paf_ticker_new ();
    *source_id = g_timeout_add (10, tick_always_true, ticker);
}

/*
 * Example PAF integration in a GMainLoop application/lib that avoids doing
 * anything if no animation is going on. It is considered that when no animation
 * takes place, an animation can only happen when an input event is triggered.
 * The timeout source will get removed automatically since paf_ticker_tick()
 * returns FALSE when it had no effect.
 */
void
timeout_removal_notify (gpointer data)
{
    MyApplicationStates *states = get_application_states ();

    states->tick_source = 0
}

void
on_input_event (MagicEvent *event, gpointer data)
{
    MyApplicationStates *states = get_application_states ();

    if (!states->tick_source && magic_event_that_triggers_animation (event))
    {
        states->tick_source = g_timeout_add_full (G_PRIORITY_DEFAULT,
                                                  10,
                                                  (GSourceFunc) paf_ticker_tick,
                                                  paf_get_default_ticker (),
                                                  timeout_removal_notify);
    }
}

void
my_application_init (MyApplicationStates *states)
{
    PafTicker *ticker = paf_ticker_new ();
}
